;. Single-cell LiIon Battery montitor program and switch enable.  File: bb1.asm
;
; Global Assembler directives
	list		c=132, n=80
	list		p=12F510
	errorlevel	-227			; Disable messages for substituting retlw 0
	errorlevel	-306			; Disable Crossing page boundary messages

#include	p12f510.inc

;
; This program controls a boost-mode converter (SMPS) powered by a single LiIon cell with a
; active high enable.  It monitors the battery for low-voltage and disables the boost
; converter when a low-battery condition is detected.  It provides a one-button, dual-
; function interface and three LED status display.  It also is capable of detecting a fault
; condition and shutting down the SMPS within approximately 10 mSec.
;
; Button Interface
;  Holding the Button for more than 3 seconds and releasing it will disable the SMPS 
;  and put the PIC to sleep.  Pressing the Button for less than 3 seconds when the PIC
;  is operating will display the current battery level on the 3 LEDs.  Pressing the Button
;  for more than 3 seconds when the PIC is asleep will enable the SMPS (Releasing the button
;  in less than 3 seconds will put the PIC asleep again without starting the SMPS).
;
; LED Display
;  Three LEDS (red/yellow/green) are used to display several operational status messages.
;    1. Power-up: The LEDs are sequenced from red -> yellow -> green on power-up
;    2. Power-down: The LEDs are sequenced from green -> yellow -> red on power-down
;    3. Battery Level: The battery level is displayed briefly when the button is pressed
;       and quickly released.  The following thresholds are displayed:
;         > 4.0 : Green LED
;         > 3.8 : Green & Yellow LEDs
;         > 3.6 : Yellow LED
;         > 3.4 : Red & Yellow LEDs
;         > 3.2 : Red LED
;    4. The Red LED is slowly blinked (on: 200 mSec; off: 1.8 sec) when the battery
;       voltage is less than 3.2 volts.
;    5. The Red LED is quickly blinked (on: 150 mSec; off: 150 mSec) for a detected 
;       fault condition.
;
;  Either blinking condition takes priority over the battery level display.
;
; Low-battery shutdown
;  The PIC power is connected to the battery supply.  The internal reference is sampled
;  by the ADC to detect the battery voltage. The SMPS power supply is disabled and the PIC
;  put to sleep when the battery voltage drops below 3.0 volts for multiple consecutive
;  samples.
;
; Fault detection
;  A PIC input may be used to detect a fault condition.  Pulling the input low (it has an
;  enabled weak pull-up) indicates a fault condition.  The SMPS is switched off when
;  a fault is detected and the Red LED rapidly blinks.  The system must be power-cycled to
;  reset the condition (e.g. Button held down to turn off and then turn back on).
;
;
; Copyright (c) 2012-2013 Dan Julio.  All rights reserved. 
;
; This program is free software: you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation, either version 3 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program.  If not, see <http://www.gnu.org/licenses/>.
;
;
; Revision History
;    9/07/13 Revision 1.0  DJJ  Initial version based on PIC16F1825 hack of bpd.
;
; ------------------------------------------------------------------------
; Processor Information:
;
; Target processor: PIC 12F510
; Processor frequency: Internal 4 MHz (1 uSec instruction cycle)
;
; Memory Layout 
;    Page 0: 0x0000 - 0x01FE	Main code and subroutines
;    Page 1: 0x0200 - 0x03FE	Subroutines
;    Page 1: 0x03FF		Factory programmed OSCCAL (must be preserved by programmer)
;
; Input/Output Pin Assignments (DIP/SOIC package)
;
;                                      Operating             Sleep
;                                  /----------------\  /----------------\
;   Pin     Signal                  Pol  I/O  def       Pol  I/O  def
;   ----------------------------------------------------------------------------
;    7 GP0  Fault Input / ICSP DATA  L    I    0 WPU     H    O    1
;    6 GP1  PSU Enable / ICSP CLK    H    O    0         H    O    0
;    5 GP2  Red Status LED Output    H    O    0         H    O    0
;    4 GP3  Button Input / VPP       L    I    1 WPU     L    I    1 WPU-WOC
;    3 GP4  Yellow Status LED Output H    O    0         H    O    0
;    2 GP5  Green Status LED Output  H    O    0         H    O    0
;
; PIC Peripheral Utilization
;   TIMER0 - used to generate an approx 10 mSec period for internal evaluations.
;   ADC - Configured to read the internal voltage reference for battery voltage computation.
;
;   REMEMBER if hacking this code: This little PIC is a bit brain-dead...
;     1. It has only 2 stack entries so no nested subroutine calls more than 2 deep.
;     2. Subroutine calls must be in the first 256 program words of each 512 word page.
;     3. Don't forget the page bit PA0 in STATUS when calling or jumping to the 2nd page. 
;     4. The last location cannot be used by program code as it stores the OSCCAL value
;        (preserved by MPLAB when programming the device).
;
; ------------------------------------------------------------------------
; File Layout: 
;
;    1. Configuration Bits
;    2. Program Constants
;    3. Macro definitions
;    4. Main code
;    5. Main code subroutines
;
;
; ------------------------------------------------------------------------
; Set the device configuration bits
;    IOSCFS = 0 (4 MHz INTOSC speed)
;    MCLRE = 0 (GP3/nMCLR pin functions as GP3)
;    nCP = 1 (Code Protection off)
;    WDTE = 0 (WDT disabled)
;    FOSC1:0 = 10 (INTOSC with 1.125 ms DRT)
;
	__config _IntRC_OSC & _WDT_OFF & _CP_OFF & _MCLRE_OFF & _IOSCFS_OFF


; ------------------------------------------------------------------------
; Program Constants
;

;
; External device bits (port number bits)
LOW_BATT_ST	equ	.2
MED_BATT_ST	equ	.4
HIGH_BATT_ST	equ	.5
FAULT_INPUT	equ	.0
PSU_EN		equ	.1
PWR_BUT		equ	.3
;
; ButFlags variable bits
KeyPress	equ	.0	; Key press just detected
TempButton	equ	.1	; Current sampled button value
LastKey		equ	.2	; Previous sample value
KeyDown		equ	.3	; Key currently pressed
KeyReleased	equ	.4	; Key release just detected
LongPress	equ	.5	; Key held down longer than long press threshold
; BUT_INIT sets the state for initial button detected on power-up from sleep
BUT_INIT	equ	((1 << KeyDown) | (1 << LastKey))
;
; DispFlags variable bits (all active true)
DoDisplayLevel	equ	.0	; Set when a valid display batt level button press is detected
DispLowBatt	equ	.1	; Set while displaying a blinking low battery condition
DispFault	equ	.2	; Set while displaying a fault condition
DispLevel	equ	.3	; Set while displaying the battery level
DispPowerUp	equ	.4	; Set while displaying the power-up animation
DispPowerDn	equ	.5	; Set while displaying the power-down animation
DispBlinkOn	equ	.7	; State of blinking LED indication (set when LED is ON)
;
; Flags variable bits (all active true)
BattCheckOk	equ	.0	; Set when enough battery samples have been collected to avg
BattLow		equ	.1	; Set for low battery indication 
BattExh		equ	.2	; Set for low battery shutdown
PowerOn		equ	.4	; Set immediately after reset to indicate a power-up
PowerUp		equ	.5	; Set to initiate a power-up sequence
PowerDown	equ	.6	; Set to initiate a power-down sequence
PowerOff	equ	.7	; Set at the end of the turn-off animation to shut down
;
; FaultFlags variable bits (all active true)
PrevFaultInput	equ	.0	; Temporary for debouncing/validating fault input
FaultDetected	equ	.1	; Set for fault detection
EnCheckFault	equ	.7	; Set after the SMPS has had a chance to start up so we
				;   don't check for a fault before the output voltage is
				;   valid
;
; TempFlags variable bits
CompTrue	equ	.0	; CompThreshold return value
SetLow		equ	.1	; Temporary used in EvalDisplay
SetMid		equ	.2	; Temporary used in EvalDisplay
SetHigh		equ	.3	; Temporary used in EvalDisplay
;
; Bank 0 Variable register equates
;
;
; Available in all banks
TempReg		equ	0x0A
TempFlags	equ	0x0B
ACCaHI		equ	0x0C	; Operand 1 hi
ACCaLO		equ	0x0D	; Operand 1 lo
ACCbHI		equ	0x0E	; Operand 2 hi / Result hi
ACCbLO		equ	0x0F	; Operand 2 lo / Result lo

; Available only in bank 0
;
Flags		equ	0x10	; Control flags
DispFlags	equ	0x11	; Flags associated with the LED display
DispAnFlags	equ	0x12	; Animation state flags (only one bit set at a time)
ButFlags	equ	0x13	; Power Button debounce flags
FaultFlags	equ	0x14	; Fault detection flags
ScanButCount	equ	0x15	; Button scan timer (counts down number of Eval intervals
				;   between button debounce scans)
ButTimer	equ	0x16	; Button press timer (incremented each button debounce scan)
BlinkTimer	equ	0x17	; LED Display indication count-down timer
BattAvgPtr	equ	0x18	; Pointer into AvgBuff (0-NUM_ADC_SAMPLES-1)
BattAvg		equ	0x19	; Average ADC value
BattLowCount	equ	0x1A	; Count of battery comparisons exceeding low threshold
BattExhCount	equ	0x1B	; Count of battery comparisons exceeding exhausted threshold

;
; Bank 1 Variable register equates
;
; AvgBuff is used to average ADC readings.  Each entry is one byte. 
;   NUM_ADC_SAMPLES must be 1, 2, 4, 8, or 16.
;   NUM_AVG_SHIFTS must be  0, 1, 2, 3, or 4   matching NUM_ADC_SAMPLES.
;
AvgBuff		equ	0x30
NUM_AVG_SAMPLES equ	.16
NUM_AVG_SHIFTS	equ	.4

;
; Initialization constants
OPTION_INIT	equ	B'10000111'	; 7: nGPWU = 1 (Wake-up On Pin Change disabled)
					; 6: nGPPU = 0 (Enable Weak Pull-Ups)
					; 5: T0CS = 0 (Timer0 Clock = Internal Instruction)
					; 4: T0SE = 0 (Timer0 inc on low-to-high)
					; 3: PSA = 0 (Prescaler assigned to Timer0)
					; 2:0: PS2:0 = 111 (256:1)

OPTION_OFF	equ	B'00000111'	; 7: nGPWU = 0 (Wake-up On Pin Change enabled)
					; 6: nGPPU = 0 (Enable Weak Pull-Ups)
					; 5: T0CS = 0 (Timer0 Clock = Internal Instruction)
					; 4: T0SE = 0 (Timer0 inc on low-to-high)
					; 3: PSA = 0 (Prescaler assigned to Timer0)
					; 2:0: PS2:0 = 111 (256:1)

GPIO_INIT	equ	B'00001000'	; No outputs enabled
GPIO_OE_INIT	equ	B'00001001'	; GP0, GP3 input

GPIO_OFF	equ	B'00001001'	; GP0 set to minimize leakage/prevent spurious wake-up
GPIO_OE_OFF	equ	B'00001000'	; GP3 input

ADCON0_INIT	equ	B'00111101'	; 7:6: ANS1:0 = 00 (No pins configured for analog in)
					; 5:4: ADCS1:0 = 11 (ADC Clock = INTOSC/4)
					; 3:2: CSH1:0 = 11 (ADC Channel = 0.6V ref)
					; 1: GO/nDONE = 0
					; 0: ADON = 1 (ADC Enabled)

ADCON0_OFF	equ	B'00111100'	; 7:6: ANS1:0 = 00 (No pins configured for analog in)
					; 5:4: ADCS1:0 = 11 (ADC Clock = INTOSC/4)
					; 3:2: CSH1:0 = 11 (ADC Channel = 0.6V ref)
					; 1: GO/nDONE = 0
					; 0: ADON = 0 (ADC Disabled)

CM1CON0_INIT	equ	B'01010011'	; 7: C1OUT = 0
					; 6: nC1OUTEN = 1 (Output NOT on C1OUT pin)
					; 5: C1POL = 0
					; 4: nC1T0CS = 1 (TMR0 Clock controlled by T0CS)
					; 3: C1ON = 0 (Comparator disabled)
					; 2: C1NREF = 0 (Negative Input = 0.6V)
					; 1: C1PREF = 1 (Positive Input = C1IN+ pin)
					; 0: nC1WU = 1 (Comparator Wake-up disabled)

;
; Global constants
EXH_BATT_THRESH		equ	.51	; round((0.6 / 3.0) * 255) [>= ==> Turn Off]
L5_BATT_THRESH		equ	.48	; round((0.6 / 3.2) * 255) [>= ==> Blink Red]
L4_BATT_THRESH		equ	.45	; round((0.6 / 3.4) * 255) [>= ==> Red]
L3_BATT_THRESH		equ	.43	; round((0.6 / 3.6) * 255) [>= ==> Red-Yellow]
L2_BATT_THRESH		equ	.40	; round((0.6 / 3.8) * 255) [>= ==> Yellow]
L1_BATT_THRESH		equ	.38	; round((0.6 / 4.0) * 255) [>= ==> Yellow-Green]
					;                          [<  ==> Green]

LOW_BATT_COUNT		equ	.100	; Required number of consecutive battery readings
					; exceeding the threshold = 1 second (100 x 10 mSec)

EXH_BATT_COUNT		equ	.100	; Required number of consecutive battery readings
					; exceeding the threshold = 1 second (100 x 10 mSec)

; Preset for an approximate 10 mSec delay in TIMER0
;  The TIMER0 prescale is set to 256:1 so that the timer increments every 256 uSec.
;  Setting TMR0 to N = 10E-3 / 256E-6 will result in a 9.984 mSec timeout (plus code
;  execution jitter of reading TMR0 = 0)
TEN_MSEC_TMR0		equ	0x100 - .39
;
; Eval interval related timer presets
BUT_COUNT_INIT		equ	.5	; 50 mSec = 5 x 10 msec eval intervals (debounce)
BUT_LONG_PRESS		equ	.60	; 3000 mSec = 60 x 50 msec debounce intervals

ANIMATE_COUNT_INIT	equ	.10	; 100 mSec = 10 x 10 msec eval intervals

BLINK_LB_ON_COUNT_INIT	equ	.20	; 200 mSec = 20 x 10 msec eval intervals
BLINK_LB_OFF_COUNT_INIT	equ	.180	; 1800 mSec = 180 x 10 msec eval intervals

BLINK_ERR_ON_COUNT_INIT	equ	.15	; 150 msec = 15 x 10 msec eval intervals
BLINK_ERR_OFF_COUNT_INIT equ	.15	; 150 msec = 15 x 10 msec eval intervals

DISP_BATT_LEVEL_INIT	equ	.50	; 500 mSec = 50 x 10 msec eval intervals


; ------------------------------------------------------------------------
; Macro Definitions
;

ENABLE_SMPS		macro
	bsf		GPIO,PSU_EN
	endm

DISABLE_SMPS	macro
	bcf		GPIO,PSU_EN
	endm

ENABLE_LOW_LED	macro
	bsf		GPIO,LOW_BATT_ST
	endm

DISABLE_LOW_LED	macro
	bcf		GPIO,LOW_BATT_ST
	endm

ENABLE_MID_LED	macro
	bsf		GPIO,MED_BATT_ST
	endm

DISABLE_MID_LED	macro
	bcf		GPIO,MED_BATT_ST
	endm

ENABLE_HIGH_LED	macro
	bsf		GPIO,HIGH_BATT_ST
	endm

DISABLE_HIGH_LED	macro
	bcf		GPIO,HIGH_BATT_ST
	endm

CHECK_THRESHOLD		macro	Thresh
	movlw		Thresh
	movwf		ACCaLO
	pagesel		CompThreshold
	call		CompThreshold
	pagesel		$
	endm


; ------------------------------------------------------------------------
; Code Block 
; 
	; Reset Vector
PAGE_0	org		0x0
	; At reset, W contains the OSCCAL value set by a MOVLW <CAL LOC>
	; as the first instruction executed.  Copy that into the OSCCAL register
	movwf		OSCCAL

	; Initialization
	call		InitSystem

	; Go back to sleep for power-on reset or wdt reset.  We only wake up
	; for an exit-from-sleep from a button press.
; ??? Can also be STATUS,GPWUF = 1
	btfsc		STATUS,NOT_PD
	call		GotoSleep

	; ----------------------------------------------------------------
	; Main loop: 
	;   - Check for fault
	;   - Check if Eval interval
	;     - Reset eval interval timer
	;     - Button sense evaluation
	;     - Battery level evaluation
	;     - Display update
	;     - Power down processing
MAIN_LOOP
	; Look for a fault condition
	call		CheckFault

	; Look for an Eval period to do some housekeeping and algorithm eval
	movf		TMR0,W
	btfss		STATUS,Z
	goto		MAIN_LOOP

	; Reset the timer for the next evaluation
	call		ResetTimer

	; We can check for faults after the first eval period after the SMSP has been
	; enabled since it should be operational within 10 mSec
	btfss		Flags,PowerOn
	bsf		FaultFlags,EnCheckFault

	; Evaluate current state
	call		EvalButton
	call		EvalBattery

	pagesel		EvalDisplay		; EvalDisplay in page 1
	call		EvalDisplay
	pagesel		$

	; Evaluate power-down (from button or low-battery condition)
	btfsc		Flags,BattExh
	call		GotoSleep
	btfsc		Flags,PowerOff
	call		GotoSleep
	goto		MAIN_LOOP



; ------------------------------------------------------------------------
; Start of Page 0 subroutines
;

; InitSystem: Power-on initialization of the processor, variables and peripherals
;
; On entry: none
;
; On exit: Processor, variables and peripherals initialized
;
InitSystem
	; Initialize the processor
	call		InitPic

	; Initialize system variables
	call		InitVars

	; Configure the timer for the initial evaluation timeout
	call		ResetTimer

	return


; InitPic: Execute a power-on initialization of the processor
;
; On entry: none
;
; On exit: PIC registers initialized
;
InitPic
	movlw		GPIO_INIT
	movwf		GPIO
	movlw		GPIO_OE_INIT
	tris		GPIO
	movlw		OPTION_INIT
	option
	movlw		CM1CON0_INIT
	movwf		CM1CON0
	movlw		ADCON0_INIT
	movwf		ADCON0
	return


; InitVars: Initialize variables
;
; On entry: Ports initialized
;
; On exit: Control variables initialized to starting conditions
;
; Notes: Assumes operation in bank 0
;
InitVars
	clrf		Flags
	bsf		Flags,PowerOn
	clrf		DispFlags
	clrf		FaultFlags
	movlw		BUT_INIT
	movwf		ButFlags
	movlw		BUT_COUNT_INIT
	movwf		ScanButCount
	movlw		BUT_LONG_PRESS
	movwf		ButTimer
	clrf		BattAvgPtr
	clrf		BattExhCount
	clrf		BattLowCount
	return


; ResetTimer: Reset TIMER0 for the next evaluation period
;
; On entry: none
;
; On exit: TMR0 preset
;
ResetTimer
	movlw		TEN_MSEC_TMR0
	movwf		TMR0
	return


; CheckFault: Look for a fault condition on the Sense input.  Disable
; the SMPS and set a flag to indicate a fault if one is detected.  Disabling the
; SMPS and setting the flag are latched and can only be cleared by power cycling.
;
; On entry: FaultFlags.EnCheckFault set if it is ok to test for a fault
;           FAULT_INPUT Sense input 
;
; On exit: SMPS enable may be cleared
;          FaultFlags.FaultDetected may be set
;
CheckFault
	; Only check if enabled to do so
	btfss		FaultFlags,EnCheckFault
	return

	; We require two consecutive fault signal detections
	btfsc		GPIO,FAULT_INPUT		; Fault Input asserted?
	goto		CHECK_FAULT_NONE		;   N: Do nothing but update state
	btfss		FaultFlags,PrevFaultInput	;   Y: Two in a row?
	goto		CHECK_FAULT_SET			;     N: Just update state
	DISABLE_SMPS					;     Y: Fault: Dispable SMPS 
	bsf		FaultFlags,FaultDetected 	;        and set flag
CHECK_FAULT_SET
	; Update state with current input
	bsf		FaultFlags,PrevFaultInput
	return

CHECK_FAULT_NONE
	bcf		FaultFlags,PrevFaultInput
	return


; EvalButton: Process button presses and update related state.  Special processing
; for the power-on button press.  Go back to sleep if the press is less than the
; long threshold, otherwise enable the SMPS.
;
; on entry: Beginning of a evaluation interval
;           ButFlags contains the current button state
;           ScanButCount contains the current button scan timer
;
; on exit: ScanButCount decremented or reset
;          ButFlags may be updated
;          Flags.PowerDown may be set
;          DispFlags.DoDisplayLevel may be set
;
EvalButton
	; Evalaute the Debounce timer
	decfsz		ScanButCount,F
	return					; not time to scan

	; Reset scan timer
	movlw		BUT_COUNT_INIT
	movwf		ScanButCount

	; Scan button to update press/release/down state
	call		ScanButton

	; Look for a new press
	btfss		ButFlags,KeyPress	; Button press just detected?
	goto		EVAL_BUTTON_TIMER	;   N: See if we're currently timing a press
	movlw		BUT_LONG_PRESS		;   Y: Initialize the long threshold timer
	movwf		ButTimer
	bcf		ButFlags,LongPress
	return

EVAL_BUTTON_TIMER
	; Look for a new release
	btfsc		ButFlags,KeyReleased	; Key Release detected this eval?
	goto		EVAL_BUTTON_SHORT	;   Y: Perform short press action if necessaty
	btfss		ButFlags,KeyDown	;   N: Button down?
	return					;     N: Do nothing
	btfsc		ButFlags,LongPress	;     Y: Timer already passed the threshold?
	return					;       N: Don't evaluate timer anymore
	decfsz		ButTimer,F		;       Y: Evalute timer: --ButTimer = 0?
	return					;         N: Continue
	bsf		ButFlags,LongPress	;         Y: Note long press threshold
	; Perform long press action
	btfss		Flags,PowerOn		; Are we powering up?
	goto		EVAL_BUTTON_POWERDOWN	;   N: Then we're powering down
	bcf		Flags,PowerOn		;   Y: Clear power up status
	ENABLE_SMPS				;      and switch on the SMPS	
	bsf		Flags,PowerUp
	return

EVAL_BUTTON_SHORT
	btfsc		ButFlags,LongPress	; Was it a long press?
	return					;   Y: Do nothing as we've done the action
	btfsc		Flags,PowerOn		; Are we powering up?
	call		GotoSleep		;   Y: Turn off again
	bsf		DispFlags,DoDisplayLevel ;   N: Normal display battery level press
	return

EVAL_BUTTON_POWERDOWN
	; Setup to turn off
	bsf		Flags,PowerDown
	return



; ScanButton : Scan the button
;
; on entry: ButFlags contains current button state
;
; on exit: ButFlags updated with new button state
;
; Note: Designed to be called every debounce interval
;
ScanButton
	; Check for button press detected
	; (we only change state on the edge a button press is detected)
	;
	; Check button status in only one location - set TempButton if
	; button depressed.
	bcf		ButFlags,TempButton
	btfss		GPIO,PWR_BUT		; sample button input
	bsf		ButFlags,TempButton	; set flag if button input = 0

	; Update KeyPress.  Assume KeyPress = 1.  Clear KeyPress if
	; any component in the equation TempButton & LastKey & !KeyDown is
	; not true.
	bsf		ButFlags,KeyPress
	btfss		ButFlags,TempButton	; Check TempButton
	bcf		ButFlags,KeyPress
	btfss		ButFlags,LastKey	; Check LastKey
	bcf		ButFlags,KeyPress
	btfsc		ButFlags,KeyDown	; Check !KeyDown
	bcf		ButFlags,KeyPress

	; Update KeyReleased.  Assume KeyReleased = 1.  Clear KeyReleased
	; if any component in the equation KeyDown & !TempButton & !LastKey is
	; not true
	bsf		ButFlags,KeyReleased
	btfss		ButFlags,KeyDown	; Check KeyDown
	bcf		ButFlags,KeyReleased
	btfsc		ButFlags,TempButton	; Check !TempButton
	bcf		ButFlags,KeyReleased
	btfsc		ButFlags,LastKey	; Check !LastKey
	bcf		ButFlags,KeyReleased

	; Update state flags
	btfss		ButFlags,TempButton	; Currently pressed?
	goto		SB1_NOT_DOWN 		;   N: set LastKey to 0
	btfsc		ButFlags,LastKey	;   Y: LastKey set?
	bsf		ButFlags,KeyDown	;      Y: set KeyDown
	bsf		ButFlags,LastKey	; LastKey = 1 since button pressed
	return

SB1_NOT_DOWN
	btfss		ButFlags,LastKey	; Lastkey already cleared?
	bcf		ButFlags,KeyDown	;   Y: clear Keydown
	bcf		ButFlags,LastKey
	return


; GotoSleep: Setup for sleep with wake-on-change on the button input
; and everything else powered down.  Put the processor to sleep.  Upon waking 
; the processor will automatically restart with a reset.
;
; On entry: none
;
; On exit: should never exit
; 
GotoSleep
	; Disable internal peripherals
	movlw		ADCON0_OFF
	movwf		ADCON0

	; Setup ports for low-power operation
	movlw		GPIO_OFF
	movwf		GPIO
	movlw		GPIO_OE_OFF
	tris		GPIO

	; Enable wake-up on change
	movlw		OPTION_OFF
	option

	; Put the processor to sleep
	sleep

	; Should never execute this statement (but simulate reset just-in-case)
	bsf		STATUS,PA0
	goto		0x03FF		; Goto the movlw OSCCAL value stored in the last
					;   location of this PIC


; EvalBattery: Evaluate the battery condition.
;  1. Read the ADC
;  2. Store the value in the averging buffer
;  3. Make an average if possible
;  4. Check for shut-off voltage condition if possible
;  5. Check for low battery condition if possible
;
; On entry: AvgBuff is the circular averaging buffer
;           BattAvgPtr points to the current entry
;           BattExhCount contains the current exhausted battery detected count
;           BattLowCount contains the current low battery detected count
;           Flags.BattCheckOk set if an average can be made
;
; On exit: AvgBuff may be updated
;          BattAvgPtr may be incremented to the next entry
;          Flags.BattCheckOk may be set
;          BattAvg may be updated
;          BattExhCount may be incremented or reset
;          Flags.BattExh may be set
;          BattLowCount may be incremented or reset
;          Flags.BattLow may be set or clear
;
EvalBattery
	; Read the ADC to get the current reading of the FVR wrt VDD
	bsf		ADCON0,GO		; Trigger conversion
	btfsc		ADCON0,GO		; Spin until conversion done
	goto		$-1

	; Store the ADC reading in the averaging array
	movlw		AvgBuff			; Setup up FSR to point to the current entry
	addwf		BattAvgPtr,W
	movwf		FSR
	; Store value
	movf		ADRES,W
	movwf		INDF
	; Increment pointer
	BANKSEL		BattAvgPtr
	incf		BattAvgPtr,F		; BattAvgPtr++
	movlw		NUM_AVG_SAMPLES
	subwf		BattAvgPtr,W		; W = ++BattAvgPtr - NUM_AVG_SAMPLES
	btfss		STATUS,Z		; NUM_AVG_SAMPLES = BattAvgPtr
	goto		EB_CHK_PTR		;   N: Continue
	clrf		BattAvgPtr		;   Y: Rollover; Set BattAvgPtr = 0
	bsf		Flags,BattCheckOk	;      Note we have enough samples to average

EB_CHK_PTR
	; See if we have enough samples to perform calculations
	btfss		Flags,BattCheckOk
	return

	; Get the average ADC value
	movlw		NUM_AVG_SAMPLES		; Loop counter
	movwf		TempReg
	movlw		AvgBuff			; Setup FSR
	movwf		FSR
	clrf		ACCbHI			; Clear accumulator
	clrf		ACCbLO
EB_AVG_LOOP
	; Sum each entry into the accumulator
	clrf		ACCaHI
	movf		INDF,W
	movwf		ACCaLO
	incf		FSR,F
	call		add16
	decfsz		TempReg,F		; Evaluate loop counter
	goto		EB_AVG_LOOP
	; Divide accumulated result appropriately to rescale back to 8-bits
	movlw		NUM_AVG_SHIFTS
	movwf		TempReg
EB_DIV_LOOP
	bcf		STATUS,C
	rrf		ACCbHI,F
	rrf		ACCbLO,F		; carry-out has most recent MSB frac value
	decfsz		TempReg,F		; Evaluate loop counter
	goto		EB_DIV_LOOP
	; Round scaled result in ACCbLO
	btfss		STATUS,C		; Last carry-out set?
	goto		EB_STORE_AVG		;   N: No rounding
	incf		ACCbLO,F		;   Y: Round up
EB_STORE_AVG
	; Save current average
	BANKSEL		BattAvg
	movf		ACCbLO,W
	movwf		BattAvg

	; ------
	; Check if the average exceeds the exhausted-battery threshold
	CHECK_THRESHOLD	EXH_BATT_THRESH
	btfsc		TempFlags,CompTrue	; Avg >= Exhausted-battery threshold?
	goto		EB_INC_BE_COUNT		;    Y: Increment BattExhCount
	clrf		BattExhCount		;    N: Clear BattExhCount
	bcf		Flags,BattExh
	goto		EB_CHK_BL

EB_INC_BE_COUNT
	; Increment the count and compare for consistent exhausted-battery detection
	incf		BattExhCount,F
	movlw		EXH_BATT_COUNT
	subwf		BattExhCount,W
	btfsc		STATUS,Z		; ++BattExhCount == EXH_BATT_COUNT?
	bsf		Flags,BattExh		;    Y: Time to shut down

EB_CHK_BL
	; -----
	; Check if the average exceeds the low-battery threshold
	CHECK_THRESHOLD	L5_BATT_THRESH
	btfsc		TempFlags,CompTrue	; Avg >= low-battery threshold?
	goto		EB_INC_BL_COUNT		;    Y: Increment BattLowCount
	clrf		BattLowCount		;    N: Clear BattLowCount
	bcf		Flags,BattLow
	return

EB_INC_BL_COUNT
	; Increment the count and compare for consistent low-battery detection
	btfsc		Flags,BattLow		; Already noted a low battery?
	return					;    Y: Don't increment the counter any more
	incf		BattLowCount,F
	movlw		LOW_BATT_COUNT
	subwf		BattLowCount,W
	btfsc		STATUS,Z		; ++BattLowCount == LOW_BATT_COUNT?
	bsf		Flags,BattLow		;    Y: Note low battery
	return 


; CompThreshold: Compare the current average against a threshold
;
; On entry: ACCaLO contains the threshold to test against
;           BattAvg contains the latest average
;
; On exit: TempFlags.CompTrue set if the average >= threshold
;
CompThreshold
	bcf		TempFlags,CompTrue	; Will be set if necessary

	; Check if the average exceeds the threshold
	movf		ACCaLO,W
	subwf		BattAvg,W		; W = BattAvg - threshold
	btfsc		STATUS,C		; C = 1 if BattAvg >= threshold
	bsf		TempFlags,CompTrue	;    C = 1: Set flag
	return


; sub16, add16: 16-bit unsigned subtraction and addition routines.  Taken from
; Microchip AN-00526.  These routines implement 16-bit unsigned operations.
;
; On entry: ACCbHI, ACCbLO contain operand 1
;           ACCaHI, ACCaLO contain operand 2
;
; On exit: ACCbHI, ACCbLO contain the result, either ACCb - ACCa or
;          ACCb + ACCa
;
; Notes: These routines are together because sub16 uses add16
;
sub16
	; Negate ACCa
	comf		ACCaLO,F
	incf		ACCaLO,F
	btfsc		STATUS,Z
	decf		ACCaHI,F
	comf		ACCaHI,F
	; Then add (ACCb + negated ACCa)
add16
	movf		ACCaLO,W
	addwf		ACCbLO,F	; add lsb
	btfsc		STATUS,C	; add in carry
	incf		ACCbHI,F
	movf		ACCaHI,W
	addwf		ACCbHI,F	; add msb
	return


; ------------------------------------------------------------------------
; Start of Page 1 subroutines
;
	org		0x0200


; EvalDisplay: Evaluate the battery display using the following priorities
;   1. Fault Detection fast blink
;   2. Power-down animation
;   3. Power-up animation
;   4. Low Battery Indication slow blink
;   5. Battery Level Display
;
; On entry: Flags.BattLow, Flags.PowerUp, Flags.PowerDown, FaultFlags.FaultDetected,
;             DispFlags.DoDisplayLevel contain state to display
;           DispFlags contains the current display state
;           BlinkTimer may contain a current count-down value
;           Flags.BattCheckOk set if BattAvg is valid
;           BattAvg contains the current battery level
;
; On exit: Battery Level display updated if necessary
;          DispFlags may be updated
;          BlinkTimer may be updated
;          Flags.PowerOff may be set
;          Flags.PowerDown may be cleared
;          Flags.PowerUp may be cleared
;
EvalDisplay
	; Priority 1: Display fault blink
	btfss		FaultFlags,FaultDetected
	goto		EVAL_DISPLAY_PRI_2
	btfsc		Flags,PowerDown			; Special case of power down while
							;   displaying fault detected?
	bsf		Flags,PowerOff			;   Y: Setup to turn off after this
							;       evaluation instead of doing
							;       the power-down animation since
							;       we're already blinking
	btfss		DispFlags,DispFault		; Already displaying FD blink?
	goto		EVAL_DISPLAY_SETUP_FD		;   N: Setup blink
	decfsz		BlinkTimer,F			;   Y: Evaluate --blink timer = 0?
	return						;      N: Do nothing
	btfss		DispFlags,DispBlinkOn		;      Y: Toggle blink and reset timer
	goto		EVAL_DISPLAY_SET_FD_ON
	DISABLE_LOW_LED
	bcf		DispFlags,DispBlinkOn
	movlw		BLINK_ERR_OFF_COUNT_INIT
	movwf		BlinkTimer
	return

EVAL_DISPLAY_SETUP_FD
	; Setup fault blink
	bsf		DispFlags,DispFault
	; Clear any existing display
	DISABLE_HIGH_LED
	DISABLE_MID_LED
	DISABLE_LOW_LED
EVAL_DISPLAY_SET_FD_ON
	; Start the blink
	ENABLE_LOW_LED
	bsf		DispFlags,DispBlinkOn
	movlw		BLINK_ERR_ON_COUNT_INIT
	movwf		BlinkTimer
	return
	

	; -----
EVAL_DISPLAY_PRI_2
	; Priority 2: Display Shutdown animation - sequence the LEDs High to Low
	btfss		Flags,PowerDown
	goto		EVAL_DISPLAY_PRI_3
	btfsc		DispFlags,DispPowerDn	; Already displaying the power-down animation?
	goto		EVAL_DISPLAY_PD		;   Y: Evaluate the display timer
	bsf		DispFlags,DispPowerDn	;   N: Setup the animation
	movlw		ANIMATE_COUNT_INIT
	movwf		BlinkTimer
	clrf		DispAnFlags
	bsf		DispAnFlags,.0		; First LED
	ENABLE_HIGH_LED
	DISABLE_MID_LED
	DISABLE_LOW_LED
	return

EVAL_DISPLAY_PD
	decfsz		BlinkTimer,F			;   Y: Evaluate --blink timer = 0?
	return						;      N: Do nothing
							;      Y: Setup next animation step
	btfsc		DispAnFlags,.0
	goto		EVAL_DISPLAY_PD_2
	btfsc		DispAnFlags,.1
	goto		EVAL_DISPLAY_PD_3
	; Final step
	DISABLE_LOW_LED
	bcf		DispFlags,DispPowerDn
	bcf		Flags,PowerDown
	bsf		Flags,PowerOff			; Ready to really power off
	return

EVAL_DISPLAY_PD_2
	movlw		ANIMATE_COUNT_INIT
	movwf		BlinkTimer
	clrf		DispAnFlags
	bsf		DispAnFlags,.1
	DISABLE_HIGH_LED
	ENABLE_MID_LED
	return

EVAL_DISPLAY_PD_3
	movlw		ANIMATE_COUNT_INIT
	movwf		BlinkTimer
	clrf		DispAnFlags
	bsf		DispAnFlags,.2
	DISABLE_MID_LED
	ENABLE_LOW_LED
	return


	; -----
EVAL_DISPLAY_PRI_3
	; Priority 3: Display start-up animation
	; -----
	bcf		DispFlags,DispFault		; Cancel the FD blink if the trigger
							;   condition goes away

	; Check for power up
	btfss		Flags,PowerUp
	goto		EVAL_DISPLAY_PRI_4
	btfsc		DispFlags,DispPowerUp	; Already displaying the power-up animation?
	goto		EVAL_DISPLAY_PU		;   Y: Evaluate the display timer
	bsf		DispFlags,DispPowerUp	;   N: Setup the animation
	movlw		ANIMATE_COUNT_INIT
	movwf		BlinkTimer
	clrf		DispAnFlags
	bsf		DispAnFlags,.0		; First LED
	DISABLE_HIGH_LED
	DISABLE_MID_LED
	ENABLE_LOW_LED
	return

EVAL_DISPLAY_PU
	decfsz		BlinkTimer,F			;   Y: Evaluate --blink timer = 0?
	return						;      N: Do nothing
							;      Y: Setup next animation step
	btfsc		DispAnFlags,.0
	goto		EVAL_DISPLAY_PU_2
	btfsc		DispAnFlags,.1
	goto		EVAL_DISPLAY_PU_3
	; Final step
	DISABLE_HIGH_LED
	bcf		DispFlags,DispPowerUp
	bcf		Flags,PowerUp			; Don't do this again
	return

EVAL_DISPLAY_PU_2
	movlw		ANIMATE_COUNT_INIT
	movwf		BlinkTimer
	clrf		DispAnFlags
	bsf		DispAnFlags,.1
	DISABLE_LOW_LED
	ENABLE_MID_LED
	return

EVAL_DISPLAY_PU_3
	movlw		ANIMATE_COUNT_INIT
	movwf		BlinkTimer
	clrf		DispAnFlags
	bsf		DispAnFlags,.2
	DISABLE_MID_LED
	ENABLE_HIGH_LED
	return


	; -----
EVAL_DISPLAY_PRI_4
	; Priority 4: Display low battery blink

	; Check for low battery condition
	btfss		Flags,BattLow
	goto		EVAL_DISPLAY_PRI_5
	btfss		DispFlags,DispLowBatt		; Already displaying LB blink?
	goto		EVAL_DISPLAY_SETUP_LB		;   N: Setup blink
	decfsz		BlinkTimer,F			;   Y: Evaluate --blink timer = 0?
	return						;      N: Do nothing
	btfss		DispFlags,DispBlinkOn		;      Y: Toggle blink and reset timer
	goto		EVAL_DISPLAY_SET_LB_ON
	DISABLE_LOW_LED
	bcf		DispFlags,DispBlinkOn
	movlw		BLINK_LB_OFF_COUNT_INIT
	movwf		BlinkTimer
	return

EVAL_DISPLAY_SETUP_LB
	; Setup low battery blink
	bsf		DispFlags,DispLowBatt
	; Clear any existing display
	DISABLE_HIGH_LED
	DISABLE_MID_LED
	DISABLE_LOW_LED
EVAL_DISPLAY_SET_LB_ON
	; Start the blink
	ENABLE_LOW_LED
	bsf		DispFlags,DispBlinkOn
	movlw		BLINK_LB_ON_COUNT_INIT
	movwf		BlinkTimer
	return
	

	; -----
EVAL_DISPLAY_PRI_5
	; Priority 5: Display battery level
	bcf		Flags,BattLow		; Cancel the LB blink if the trigger
						;   condition goes away

	; Make sure there is a valid average
	btfss		Flags,BattCheckOk
	goto		EVAL_DISP_CLR_DISPLAY

	; Check if we are currently displaying the battery level
	btfsc		DispFlags,DispLevel
	goto		EVAL_DISP_CHK_BL_TMR

	; Check if we were commanded to display the level
	btfss		DispFlags,DoDisplayLevel
	goto		EVAL_DISP_CLR_DISPLAY
	bcf		DispFlags,DoDisplayLevel
	bsf		DispFlags,DispLevel
	movlw		DISP_BATT_LEVEL_INIT
	movwf		BlinkTimer
	goto		EVAL_DISP_SHOW_LEVEL

EVAL_DISP_CHK_BL_TMR
	; Check if the timer has expired
	decfsz		BlinkTimer,F		; --Timer = 0?
	goto		EVAL_DISP_SHOW_LEVEL	;   N: Continue to update the level display
	bcf		DispFlags,DispLevel	;   Y: Stop the display
	goto		EVAL_DISP_CLR_DISPLAY	;      and clear the LEDs


EVAL_DISP_SHOW_LEVEL
	; Evaluate current battery level and display
	bcf		TempFlags,SetLow	; Will be set if necessary
	bcf		TempFlags,SetMid
	bcf		TempFlags,SetHigh

	CHECK_THRESHOLD	L4_BATT_THRESH
	btfss		TempFlags,CompTrue	; Avg >= L4?
	goto		EVAL_DISP_L3		;   N: Check next lower threshold
	bsf		TempFlags,SetLow	;   Y: Set only Low LED
	goto		EVAL_DISP_UPD_DISPLAY

EVAL_DISP_L3
	CHECK_THRESHOLD	L3_BATT_THRESH
	btfss		TempFlags,CompTrue	; Avg >= L3?
	goto		EVAL_DISP_L2		;   N: Check next lower threshold
	bsf		TempFlags,SetLow	;   Y: Set Low and Mid LEDs
	bsf		TempFlags,SetMid
	goto		EVAL_DISP_UPD_DISPLAY

EVAL_DISP_L2
	CHECK_THRESHOLD	L2_BATT_THRESH
	btfss		TempFlags,CompTrue	; Avg >= L2?
	goto		EVAL_DISP_L1		;   N: Check next lower threshold
	bsf		TempFlags,SetMid	;   Y: Set only Mid LED
	goto		EVAL_DISP_UPD_DISPLAY

EVAL_DISP_L1
	CHECK_THRESHOLD	L1_BATT_THRESH
	btfss		TempFlags,CompTrue	; Avg >= L1?
	goto		EVAL_DISP_SET_HIGH	;   N: Must be full charge
	bsf		TempFlags,SetMid	;   Y: Set Mid and High LEDs
	bsf		TempFlags,SetHigh
	goto		EVAL_DISP_UPD_DISPLAY

EVAL_DISP_SET_HIGH
	bsf		TempFlags,SetHigh

EVAL_DISP_UPD_DISPLAY
	btfsc		TempFlags,SetLow
	ENABLE_LOW_LED
	btfss		TempFlags,SetLow
	DISABLE_LOW_LED

	btfsc		TempFlags,SetMid
	ENABLE_MID_LED
	btfss		TempFlags,SetMid
	DISABLE_MID_LED

	btfsc		TempFlags,SetHigh
	ENABLE_HIGH_LED
	btfss		TempFlags,SetHigh
	DISABLE_HIGH_LED
	return

EVAL_DISP_CLR_DISPLAY
	; Clear display
	DISABLE_HIGH_LED
	DISABLE_MID_LED
	DISABLE_LOW_LED
	return


	end

